#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <iostream.h>

#include <standard.h>
#include <qtree.h>


//----------------------------------------------------------------------
// NodeCount: return the number of nodes inthe given subtree
//----------------------------------------------------------------------
int QPage_t::NodeCount(int rootIndex)
{
  if (rootIndex == SINil)
    return 0;
  if (slot[rootIndex].isleaf)
    return 1;
  int count = 1;
  for (int q = 0; q < 4; q++)
    if (slot[rootIndex].ChildType(q) != PTExternal)
      count += NodeCount(slot[rootIndex].child[q]);
  return count;
}

//----------------------------------------------------------------------
// MoveTree: moves the tree rooted at ri in p1, to p2. Frees space in p1.
// returns the root of the new tree in p2
//----------------------------------------------------------------------
int MoveTree(int ri, QPage_t *p1, QPage_t *p2)
{
  if (ri == SINil)
    return SINil;
  int si = p2->AllocSlot();
  Assert(si != SINil);
  p2->slot[si] = p1->slot[ri];
  if (!p1->slot[ri].isleaf)
  {
    for (int q = 0; q < 4; q++)
    {
      if (p1->slot[ri].ChildType(q) == PTInternal)
        p2->slot[si].child[q] = MoveTree(p1->slot[ri].child[q], p1, p2);
    }
  }
  p1->FreeSlot(ri);
  return si;
}

//----------------------------------------------------------------------
// Traverse: traverses the path from the root <qroot> along point given
// by pt, assuming that the root represents an extent *pExtent, as long as
// the path remains within the page. Returns the extent of the last node
// in pExtent, and the slot corresponding to the last node as the return
// value 
//----------------------------------------------------------------------
int QPage_t::Traverse(int qroot, Rect_t *pExtent, const Point_t &pt)
{
  int si = root[qroot];
  int q;
  while (1)
  {
    if (slot[si].isleaf)
      return si;
    q = pExtent->QuadrantOf(pt);
    if ((slot[si].child[q] == SINil) || 
        (slot[si].ChildType(q) == PTExternal))
      return si;
    si = slot[si].child[q];
    pExtent->Shrink(q);
  }
  // NOTREACHED
  Assert(False);
  return -1;
}

//----------------------------------------------------------------------
// BuildPath: build the path from the root till the node where to insert
// returns the numbers of pages in the path and the extent corresponding
// to the leaf, in *pLeafExtent
//----------------------------------------------------------------------
int QTree_t::BuildPath(PathEntry_t path[], const Point_t &pt)
{
  int level = -1;
  int pageIndex = rootPage;
  Rect_t curExtent;
  int qroot = 0;
  
  curExtent = extent;
  while (1)
  {
    level += 1;
    path[level].pageIndex = pageIndex;
    path[level].page = disk.FetchPage(pageIndex);
    path[level].qroot = qroot;
    path[level].rootExtent = curExtent;
    path[level].leafIndex = path[level].page->Traverse(qroot, &curExtent, pt);
    path[level].leafExtent = curExtent;
    
    Slot_t &slot = path[level].page->slot[path[level].leafIndex];
    qroot = curExtent.QuadrantOf(pt);
    if (slot.isleaf || (slot.child[qroot] == SINil))
    {
      break;
    }
    else
    {
      Assert(slot.ChildType(qroot) != PTInternal);
      pageIndex = slot.child[qroot];
      curExtent.Shrink(qroot);
    }
  }
  return level;
}

//----------------------------------------------------------------------
// Partition: partition the set of four numbers as best as possible
// returns a bitmap representation of the larger sets of the partition
//----------------------------------------------------------------------
unsigned int Partition(int n[])
{
  int sum = 0;
  for (int i = 0; i < 4; sum += n[i++]);
  int half = sum / 2;
  int curBest = sum; 
  unsigned part;
  for (i = 1; i < 15; i++)
  {
    for (int j = 0, sum = 0; j < 4; j++)
      sum += ((i >> j) & 1)? n[j]:0;
    if (sum >= half)
      if ((sum - half) < curBest)
      {
        part = i;
        curBest = sum - half;
      }
  }
  return part;
}

//----------------------------------------------------------------------
// RemoveRoot: removes the root of the tree from the given page. (as
// specified in the pathentry. returns the root. Updates pathentry
//----------------------------------------------------------------------
void RemoveRoot(PathEntry_t *pe, Slot_t *proot, const Point_t &pt)
{
  // cout << "M";
  QPage_t *page = pe->page;
  Assert(page->TreeCount() == 1);
  int rootIndex = page->root[pe->qroot];
  Slot_t &rootSlot = page->slot[rootIndex];
  Assert(!rootSlot.isleaf);
  proot->inuse = True;
  proot->isleaf = False;
  for (int q = 0; q < 4; q++)
  {
    proot->SetChildType(q, PTExternal);
    if (rootSlot.ChildType(q) == PTExternal)
    {
      page->root[q] = SINil;
      proot->child[q] = rootSlot.child[q];
    }
    else
    {
      page->root[q] = rootSlot.child[q];
      if (page->root[q] == SINil)
        proot->child[q] = SINil;
      else
        proot->child[q] = pe->pageIndex;
    }
  }
  pe->qroot = pe->rootExtent.Shrink(pt);
  page->FreeSlot(rootIndex);
}

//----------------------------------------------------------------------
// SplitPage: splits the given page (as specified through pe.) proot is
// the slot that points to this page. So SplitPage updates proot
// appropriately.
//----------------------------------------------------------------------
void SplitPage(PathEntry_t *pe, Slot_t *proot, const Point_t &pt)
{
  // cout << "S";
  int page1Index = pe->pageIndex;
  QPage_t *page1 = pe->page;
  Assert(page1->TreeCount() > 1);
  int page2Index = disk.AllocPage();
  if (page2Index == PINil)
  {
    cerr << "Insert: Disk full !" << endl;
    exit(-1);
  }
  QPage_t *page2 = disk.FetchPage(page2Index);
  
  int treeSize[4];
  for (int q = 0; q < 4; q++)
    treeSize[q] = page1->NodeCount(page1->root[q]); 
  int part = Partition(treeSize);
  // move the trees out
  for (q = 0; q < 4; q++)
  {
    if (!((part >> q) & 1))
    {
      page2->root[q] = MoveTree(page1->root[q], page1, page2);
      page1->root[q] = SINil;
    }
  }
  
  for (q = 0; q < 4; q++)
  {
    if ((proot->ChildType(q) == PTExternal) && 
        (proot->child[q] == page1Index) &&
        (page1->root[q] == SINil))      // this tree moved to different page
      proot->child[q] = page2Index;
  }
  if (page1->root[pe->qroot] == SINil)  // root has move into page2
  {
    Assert(page2->root[pe->qroot] != SINil);
    pe->pageIndex = page2Index;
    pe->page = page2;
    pe->leafExtent = pe->rootExtent;
    pe->leafIndex = page2->Traverse(pe->qroot, &pe->leafExtent, pt);
  }
}

//----------------------------------------------------------------------
// InsertNode: inserts one node into the tree, as the qth child of the
// leaf given by path[level].leafIndex. May cause splits and root
// motions. Updates path consistently. 
//----------------------------------------------------------------------
void InsertNode(PathEntry_t path[], Slot_t slot, int q, const Point_t pt,
                int *pLevel, int *pRootPage)
{
  int nextq;
  Slot_t nextSlot;
  Boolean haveSlot = True;
  Rect_t treeExtent = path[0].rootExtent; // saved value
  
  for (int level = *pLevel; (level >= 0) && haveSlot; level -= 1)
  {
    haveSlot = False;
    int si = path[level].page->AllocSlot();
    if (si == SINil)                    // page is filled
    {
      if (path[level].page->TreeCount() == 1)
      {
        nextq = path[level].qroot;
        int rootIndex = path[level].page->root[path[level].qroot];
        if (rootIndex == path[level].leafIndex)
          path[level].leafIndex = SINil; // indicate that the leaf moved
        RemoveRoot(&path[level], &nextSlot, pt);
        haveSlot = True;
      }
      else
      {
        Assert(level != 0);             // root page has only 1 root in it
        SplitPage(&path[level],
                  &path[level-1].page->slot[path[level-1].leafIndex],
                  pt);
      }
      si = path[level].page->AllocSlot();
      Assert(si != SINil);
    }
    // at this point, si contains a new slot.
    path[level].page->slot[si] = slot;
    if (path[level].leafIndex == SINil) // i.e. the leaf moved
    {
      nextSlot.SetChildType(q, PTExternal);
      nextSlot.child[q] = path[level].pageIndex;
      path[level].page->root[q] = si;
      int q1 = path[level].leafExtent.QuadrantOf(pt);
      if (q1 == q)                      // move leafIndex to si
      {
        path[level].leafIndex = si;
        path[level].leafExtent.Shrink(q1);
      }
      else                              // this page is no longer in the path?
      {
        // it is not clear how this case can occur. If it occurs, this is
        // what should be done about it. But it can probably not occur.
        cerr << "Weird case!!" << endl;
        *pLevel -= 1;
        for (int l = level; l < *pLevel; path[l] = path[l+1], l++);
      }
    }
    else
    {
      Slot_t &leafSlot = path[level].page->slot[path[level].leafIndex];
      leafSlot.SetChildType(q, PTInternal);
      leafSlot.child[q] = si;
      int q1 = path[level].leafExtent.QuadrantOf(pt);
      if (q1 == q)                      // move leafIndex to si
      {
        path[level].leafIndex = si;
        path[level].leafExtent.Shrink(q1);
      }
    }
    slot = nextSlot;
    q = nextq;
  } // end of (for each level in the path)
  if (haveSlot)                     // root page overflows
  {
    int pageIndex = disk.AllocPage();
    if (pageIndex == PINil) 
    {
      cerr << "Insert: Disk full" << endl;
      exit(-1);
    }
    QPage_t *page = disk.FetchPage(pageIndex);
    int si = page->AllocSlot();
    Assert(si != SINil);
    *pRootPage = pageIndex;
    page->root[0] = si;
    page->slot[si] = slot;
    *pLevel += 1;
    for (level = *pLevel; level > 0; path[level] = path[level-1], level--);
    path[0].page = page;
    path[0].pageIndex = pageIndex;
    path[0].qroot = 0;
    path[0].rootExtent = treeExtent;
    path[0].leafIndex = si;
    path[0].leafExtent = treeExtent;
  }
}

//----------------------------------------------------------------------
Boolean QTree_t::Insert(const Point_t &pt, Data_t *data)
{
  Assert(extent.Contains(pt));
  if (rootPage == PINil)
  {
    if ((rootPage = disk.AllocPage()) == PINil)
    {
      cerr << "Insert: Disk full!" << endl;
      return False;
    }
    QPage_t *page = disk.FetchPage(rootPage);
    int si = page->AllocSlot();
    Assert(si != SINil);                   // we just allocated a new page
    page->root[0] = si;
    page->slot[si].isleaf = True;
    page->slot[si].l.data = data;
    page->slot[si].l.x = pt.x;
    page->slot[si].l.y = pt.y;
    // cout << "Inserting " << pt.x << ", " << pt.y << ": ";
    // cout << endl;
    return True;
  }

  // insert in non empty tree
  const int MaxPath = 32;
  PathEntry_t path[MaxPath];
  int level = BuildPath(path, pt);
  QPage_t *page = path[level].page;
  int si = path[level].leafIndex;

  // check if data already exists at point
  if (page->slot[si].isleaf && (page->slot[si].l.x == pt.x) &&
      (page->slot[si].l.y == pt.y))
  {
    return False;
  }
  // cout << "Inserting " << pt.x << ", " << pt.y << ": ";
  while (1)
  {
    Slot_t newSlot;
    Slot_t &leafSlot = path[level].page->slot[path[level].leafIndex];
    Rect_t &leafExtent = path[level].leafExtent;
    if (leafSlot.isleaf)
    {
      newSlot = leafSlot;
      leafSlot.isleaf = False;
      for (int i = 0; i < 4; leafSlot.child[i++] = SINil);
      int q = leafExtent.QuadrantOf(Point_t(newSlot.l.x, newSlot.l.y));
      InsertNode(path, newSlot, q, pt, &level, &rootPage);
    }
    else
    {
      newSlot.inuse = True;
      newSlot.isleaf = True;
      newSlot.l.x = pt.x, newSlot.l.y = pt.y;
      newSlot.l.data = data;
      int q = leafExtent.QuadrantOf(pt);
      Assert(leafSlot.child[q] == SINil);
      InsertNode(path, newSlot, q, pt, &level, &rootPage);
      break;
    }
  }
  // cout << endl;
  return True;
}
//-----------------------------------------------------------------------------
// COMPACTION CODE: KARTHIK
//-----------------------------------------------------------------------------
int CopyTreeToDisk( QPage_t& page, QTNode_t* node)
{
  if ( node == NULL )
    return -1 ; 

  // Copy this node to disk
  int newSlot = page.AllocSlot() ;

  if ( node->isleaf )
  {
    page.slot[newSlot].isleaf = True ;

    page.slot[newSlot].l.x = node->l.x ;
    page.slot[newSlot].l.y = node->l.y ;
     
    // WATCH OUT THE DATA PTR IS NOT BEING COPIED HERE
    page.slot[newSlot].l.data = NULL ;  

    delete node ; 
    return newSlot ; 
  }

  page.slot[newSlot].isleaf = False ;
    
  /*
  if ( node->ChildType() == PTExternal )
  {
    for ( int q = 0 ; q < 4; q++ )
    {
      page.slot[newSlot].SetChildType(q,PTExternal) ;
      page.slot[newSlot].child[q] = node->page[q] ; 
    } 

    delete node ; 
    return newSlot ; 
  } 

  // Intermediate node 
  for ( int r = 0 ; r < 4 ; r++ )
  {
    page.slot[newSlot].child[r] = CopyTreeToDisk( page, node->child[r] ) ;
    page.slot[newSlot].SetChildType(r,PTInternal) ;

    // Since the memory has been deallocated set the child pointer to null
    node->child[r] = NULL ; 
  }
  */

  for ( int r = 0 ; r < 4 ; r++ )
  {
    if ( node->ChildType(r) == PTInternal )
    {
      page.slot[newSlot].child[r] = CopyTreeToDisk( page, node->child[r] ) ;
      page.slot[newSlot].SetChildType(r,PTInternal) ;
    }
    else if ( node->ChildType(r) == PTExternal )
    {
      page.slot[newSlot].SetChildType(r,PTExternal) ;
      page.slot[newSlot].child[r] = node->page[r] ; 
    }
  }
      
  delete node ;
  return newSlot ; 
}
    

//-----------------------------------------------------------------------------
void CopyInMemoryTreeToDiskPage (QPage_t *page, QTNode_t* root, int rootNumber)
{
  if ( root == NULL )
    return ;
  Assert( page != NULL ) ;
  page->root[rootNumber] = CopyTreeToDisk( *page, root ) ;
}
      
//-----------------------------------------------------------------------------
int Partition( int [], int [], int ) ;// The kartikayean partition
int RecCompact(QPage_t& page, int rootIndex, QTNode_t*& qtPtr, 
			      int& pathLength, int M)
{
  void sort(int[],int[]) ;

  if ( rootIndex == -1 )
    return 0 ;
  Slot_t &rootSlot = page.slot[rootIndex];
  QTNode_t* qtNew = new QTNode_t ;  // Create a new in-memory node
  Assert( qtNew != NULL ) ;

  if (rootSlot.isleaf)
  {
    // Copy the leaf to memory 
    qtNew->isleaf = True ;
    qtNew->l.x = rootSlot.l.x ;
    qtNew->l.y = rootSlot.l.y ;

    // WATCH OUT THE DATA PTR IS NOT BEING COPIED HERE
    qtNew->l.data = NULL ;  

    // Free the slot occupied in the page by this leaf
    page.FreeSlot(rootIndex);

    // Propagate this node to its parent
    qtPtr = qtNew ; 

    // Return the path length of the leaf as 0	
    pathLength = 0 ; 

    // Return to the parent saying that the number of nodes is 1
    return 1 ; 
  }

  int childSize[4] = {0,0,0,0}; 
  int childPath[4] = {-1,-1,-1,-1} ; 	 

  int numNodes = 0 ;  

  QTNode_t* qtChildNodes[4] = { NULL, NULL, NULL, NULL } ;  

  // New node is not going to be a leaf any more
  qtNew->isleaf = False ;

  // Traverse each of the child of this node calling RecCompact 
  for (int q = 0; q < 4; q++)
  {
    if ( rootSlot.child[q] == -1 )
    {
      qtChildNodes[q] = NULL ; 
      childSize[q] = 0;
    }
    else if (rootSlot.ChildType(q) == PTInternal)
    {
      childSize[q] = RecCompact(page, rootSlot.child[q], qtChildNodes[q], 
				childPath[q], M);
    }
    else if ( rootSlot.ChildType(q) == PTExternal )
    {
      // Check really if an external child exists 
      if ( rootSlot.child[q] >= 0 )
      {
        QPage_t* p = disk.FetchPage(rootSlot.child[q]);
        Assert( p != NULL ) ;
        childSize[q] = RecCompact(*p, p->root[q], qtChildNodes[q], 
				  childPath[q], M);
        if (p->usedCount == 0) 
          disk.FreePage(rootSlot.child[q]);
      } 
    }
  } // end for 
    

  // Compute the total number of nodes of all subtrees of this node 
  for ( q = 0 ; q < 4 ; numNodes += childSize[q++] ) ;    

  int max[4] ;

  // Sort the trees in increasing order of path length 
  sort(childPath,max) ;

  // Check if the total number of nodes of all subtrees are less
  // than the minimum bucket occupancy
  if ( numNodes < M )
  {
    // There is enough room to accomodate the sub trees and this node  
    // Set all the subtrees to be in memory. PTInternal means they are in memory 
    for ( q = 0 ; q < 4 ; qtNew->SetChildType(q,PTInternal),q++ ) ;   
    
    // qtNew->SetChildType(PTInternal) ;   

    // Represent this node in memory and connect the child pointers
    for ( q = 0 ; q < 4 ; qtNew->child[q] = qtChildNodes[q++] ) ;  

    qtPtr = qtNew ;

    // Free the slot occupied in the page by this node
    page.FreeSlot(rootIndex);

    // return the maximum path length of the subtrees
    pathLength = childPath[max[3]] ;

    return numNodes + 1;            // Return this node and the number
                                    // of nodes under this subtree 
  }

  if (  numNodes >= M )
  {
    // There is no room to accomodate this node 

    int assign[4] = { 0, 0, 0, 0 } ;

    int decide[4] = { 0, 0, 0, 0 } ;
    const int PUT_DOWN = 1;
    const int PROPAGATE = 2 ;

    decide[max[0]] = PUT_DOWN ;	

    for ( int i = 1 ; i < 4 ; i++ )
    {
       if (childPath[max[i-1]] == childPath[max[i]]) 
       {
	  decide[max[i]] = PUT_DOWN ; 			// Drop the subtree  
       } 		 
       else
       {
          int total = 0 ;
	  for ( int j = i ; j < 4 ; total += childSize[max[j++]]) ;

	  if ( total < M )
	  {
	    for ( j = i ; j < 4 ; decide[max[j]] = PROPAGATE, j++ ) ;
            break ;
          }
	  else
	    decide[max[i]] = PUT_DOWN ; 	
       }
    }

    int propChildSize[4] = { 0, 0, 0, 0 } ;
    int putChildSize[4] = { 0, 0, 0, 0 } ;
    int propTotal = 0 ;
    int putTotal = 0 ; 
    int putMax = -1 ; 

    for ( i = 0 ; i < 4 ; i++ )
    {
      propChildSize[i] = decide[i] == PROPAGATE ? childSize[i] : 0 ;
      propTotal += propChildSize[i] ;

      putChildSize[i] =  decide[i] == PUT_DOWN ? childSize[i] : 0 ;
      putTotal += putChildSize[i] ;

      int tmp = (decide[i] == PUT_DOWN) + childPath[i];
      if ( putMax < tmp)
       	putMax = tmp;
    }

    // Call Partition function that assigns the put down trees to pages optimally
    int reqPages = Partition(putChildSize,assign,QPage_t::SlotsPerPage) ;
//  int reqPages = Partition(putChildSize,assign,M) ;

  cout.form("Decide: %d\t%d\t%d\t%d\t",decide[0],decide[1],decide[2],decide[3]) ; 
  cout.form("ChildSize: %d\t%d\t%d\t%d\t",childSize[0],childSize[1],childSize[2],childSize[3]) ; 
  cout.form("ChildPath: %d\t%d\t%d\t%d\n",childPath[0],childPath[1],childPath[2],childPath[3]) ; 
//  cout.form("Partition: %d\t%d\t%d\t%d\n",assign[0],assign[1],assign[2],assign[3]) ; 

//     cout << "Required: " << reqPages << " ";
    
    int newPages[4] = { -1, -1 ,-1, -1 } ;

    for ( int q = 0 ; q < 4 ; qtNew->page[q++] = -1 ) ;
    
    for ( q = 0 ; q < reqPages ; q++ )
    {
      // Allocate new page
      newPages[q] = disk.AllocPage() ;
      Assert( newPages[q] != -1) ;
      QPage_t *page = disk.FetchPage(newPages[q]);

      // flush out the nodes of the tree in memory to the current
      // iterating disk page
      for ( int tree = 0 ; tree < 4 ; tree++ )
      {
        // Check to which page the tree should go and make sure that the
        // tree is non empty
        if ( decide[tree] == PUT_DOWN && (assign[q] & ( 1 << tree )) 
	     && qtChildNodes[tree] != NULL )
        {
          // Watch out the memory of in-memory tree is deleted once the 
          // node is copied to disk

          CopyInMemoryTreeToDiskPage(page, qtChildNodes[tree], tree ) ;
          qtNew->page[tree] = newPages[q] ;     

          // Since the memory of qtChildNodes[tree] has been deallocated
          // set it to NULL
          qtChildNodes[tree] = NULL ; 

	  qtNew->SetChildType(tree,PTExternal) ;
        } 
      } 
    }
    
    // Free the slot occupied in the page by this node
    page.FreeSlot(rootIndex);

    // Set the appropriate pointers for the trees to move up
    for ( q = 0 ; q < 4 ; q++ )
    {
      if ( decide[q] == PROPAGATE )
      {
	qtNew->SetChildType(q,PTInternal) ;
	qtNew->child[q] = qtChildNodes[q] ;
      }
    }

    qtPtr = qtNew ; 

    // Return the path length. 
    pathLength = putMax;

    return propTotal + 1 ; 
  }

  // Should not be reached
  Assert(False) ;

  return 0 ;
}
		
//-----------------------------------------------------------------------------
void QTree_t::Compact(int M)
{
  Assert(M <= QPage_t::SlotsPerPage);
  if (rootPage == -1) 
    return;
  QPage_t* curRootPage = disk.FetchPage(rootPage);
  QTNode_t* memoryRootPtr = NULL ; 
  int path ; 
  int numNodes = RecCompact(*curRootPage, curRootPage->root[0], 
			    memoryRootPtr, path, M);

  Assert( numNodes <= M ) ;	 

  // Copy the remaining part of the in-memory tree to disk 
  int newRootPage = disk.AllocPage() ;
  Assert(newRootPage != -1);
  
  CopyInMemoryTreeToDiskPage(disk.FetchPage(newRootPage), memoryRootPtr, 0 ) ;

  Assert(curRootPage->usedCount == 0);

  disk.FreePage(rootPage);
  rootPage = newRootPage ;
}


//----------------------------------------------------------------------
Data_t *QTree_t::Remove(const Point_t &pt)
{
  return NULL;
}

//----------------------------------------------------------------------
Data_t *QTree_t::Search(const Point_t &pt)
{
  return NULL;
}

//----------------------------------------------------------------------
void QTree_t::RecTraverse(QPage_t *page, int si, LeafOp_t fn, 
                          int level, int pageLevel)
{
  if (si == SINil)
    return;
  if (page->slot[si].isleaf)
  {
    fn(page->slot[si].l, level, pageLevel);
    return;
  }
  for (int q = 0; q < 4; q++)
  {
    if (page->slot[si].child[q] == SINil)
      continue;
    if (page->slot[si].ChildType(q) == PTExternal)
    {
      QPage_t *p = disk.FetchPage(page->slot[si].child[q]);
      RecTraverse(p, p->root[q], fn, level+1, pageLevel+1);
    }
    else
    {
      RecTraverse(page, page->slot[si].child[q], fn, level+1, pageLevel);
    }
  }
}
  
//----------------------------------------------------------------------
void QTree_t::Traverse(LeafOp_t fn)
{
  if (rootPage == PINil)
  {
    return;
  }
  QPage_t *page = disk.FetchPage(rootPage);
  RecTraverse(page, page->root[0], fn, 1, 1);
}

//----------------------------------------------------------------------
static void PrintLeaf(Leaf_t &l, int nodeLevel, int pageLevel)
{
  cout << "[" << l.x << ", " << l.y << ", " << l.data << "]"
       << "[" << nodeLevel << "][" << pageLevel << "]" << endl;
}

//----------------------------------------------------------------------
void QTree_t::Print()
{
  if (rootPage == PINil)
    cout << "<NULL TREE>" << endl;
  else
    Traverse(PrintLeaf);
}

//----------------------------------------------------------------------
static int countAccesses;
static int countLeaves;

static void CountStats(Leaf_t &l, int nodeLevel, int pageLevel)
{
  countAccesses += pageLevel;
  countLeaves += 1;
}
//----------------------------------------------------------------------
void QTree_t::PrintStats()
{
  countAccesses = 0;
  countLeaves = 0;
  Traverse(CountStats);
#ifdef VERBOSE
  cerr << "Mean Page Access: " << (double)countAccesses/countLeaves << endl; 
#else
  cerr.form("%6.4lf", (double)countAccesses/countLeaves);
#endif
}
//----------------------------------------------------------------------
void Disk_t::PrintStats()
{
  const int NParts = 16;
  int nPages[NParts];
  int nodeCount = 0;
  int pageCount = 0;
  const int SPP = QPage_t::SlotsPerPage;
  
  for (int i = 0; i < NParts; nPages[i++] = 0);
  for (i = 0; i < MaxNoOfPages; i++)
  {
    if (pageArray[i] != NULL)
    {
      pageCount += 1;
      nodeCount += pageArray[i]->usedCount;
      int j = (pageArray[i]->usedCount * NParts)/(SPP+1);
      nPages[j] += 1;
    }
  }
#ifdef VERBOSE
  cerr << "Number of nodes: " << nodeCount << endl;
  cerr << "Number of pages: " << pageCount << endl;
  cerr << "Average occupancy: " << (double)nodeCount*100/(pageCount*SPP) 
       << "%" << endl;
  cerr << "Histogram" << endl;
  for (i = 0; i < NParts; i++)
  {
    cerr << (i*(SPP+1)/NParts) << " <= occupancy < " << ((i+1)*(SPP+1)/NParts)
         << ": " << nPages[i] << endl;
  }
#else
  cerr.form("%5d\t%5d\t%6.4lf", nodeCount, pageCount, 
            (double)nodeCount*100/(pageCount*SPP));
#ifdef HIST 
  for (i = 0; i < NParts; i++)
    cerr.form("\t%d", nPages[i]);
#endif
#endif
}
//----------------------------------------------------------------------
void InsertAll(QTree_t *t, int step)
{
  for (int i = 0; i < t->extent.p2.x; i+=step)
  {
    for (int j = 0; j < t->extent.p2.y; j+=step)
    {
      t->Insert(Point_t(i,j), NULL);
    }
  }
}
//----------------------------------------------------------------------
void InsertDiagonal(QTree_t *t, int step)
{
  for (int i = 0; i < t->extent.p2.x; i+=step)
  {
    t->Insert(Point_t(i,i), NULL);
  }
}

//----------------------------------------------------------------------
extern "C" long random(void);
extern "C" int srandom(int);
//----------------------------------------------------------------------
void InsertRandomDiagonal(QTree_t *t, int count)
{
  int i = 0;
  while (i < count)
  {
    int x = random()%t->extent.p2.x;
    if (t->Insert(Point_t(x,x), NULL))
      i += 1;
  }
}     

//----------------------------------------------------------------------
void InsertRandom(QTree_t *t, int count)
{
  int i = 0;
  while (i < count)
  {
    Point_t pt(random()%t->extent.p2.x, random()%t->extent.p2.y);
    if (t->Insert(pt, NULL))
      i += 1;
  }
}

//----------------------------------------------------------------------
const int influence = 1024;
void InsertSkewed(QTree_t *t, int count, Point_t corner[], int size)
{
  int i = 0;
  Point_t pt;
  while (i < count)
  {
    int j = random() % (size+1);
    if (j == size)                      // put on background
    {
      pt = Point_t(random()%t->extent.p2.x, random()%t->extent.p2.y);
    }
    else
    {
      pt.x = corner[j].x + random()%influence;
      pt.y = corner[j].y + random()%influence;
    }
    if (t->Insert(pt, NULL))
      i += 1;
  }
}

//----------------------------------------------------------------------
Disk_t disk;
QTree_t t;

const int NCorners = 32;
int main(int argc, char *argv[])
{
  Assert(argc == 3);
  srandom ( atoi(argv[2]) ) ; 
  int numInsert = atoi(argv[1]);
  t.extent = Rect_t(0,0,xWorldMax,yWorldMax);
//InsertDiagonal(&t, (1<<17));
//  InsertRandomDiagonal(&t, numInsert);
#ifdef SKEWED
  Point_t corner[NCorners];
  for (int i = 0; i < NCorners; i++)
    corner[i] = Point_t(i*65536, i*65536);
  InsertSkewed(&t, numInsert, corner, NCorners);
#endif
 
 
 InsertRandom(&t, numInsert);
 // for ( int i = 2000 ; i < 200000 ; i += i/4 ) 
 // {
  // InsertRandom(&t, i/4 );
  //cerr << "63\t00\t00\t00\t" << i << endl ;  
#ifdef VERBOSE
  t.PrintStats();
  disk.PrintStats();
#else
  cerr << "0.000\t";
  t.PrintStats();
  cerr << "\t";
  disk.PrintStats();
  cerr << endl;
#endif
    
  // for (int m = 132; m < QPage_t::SlotsPerPage; m+=1)
  for (int m = 40; m < 204 ;  m+=1)
  {
    t.Compact(m);
#ifdef VERBOSE
    cerr << "M=" << m << " ------------------------------------------"<<endl;
    t.PrintStats();
    disk.PrintStats();
#else
    cerr.form("%6.4lf\t", (double)m*100/QPage_t::SlotsPerPage);
    t.PrintStats();
    cerr << "\t";
    disk.PrintStats();
    cerr << endl;
#endif
  }
  //}
}
