/* this may look like c it is really -*-c++-*- */
#ifndef __QPAGE_H
#define __QPAGE_H

typedef unsigned XDim_t;
typedef unsigned YDim_t;

//----------------------------------------------------------------------
// xWorldMin <= x < xWorldMax
// yWorldMin <= y < yWorldMax
// Datatype Dim *MUST* be able to store the WorldMax value
//----------------------------------------------------------------------
const XDim_t xWorldMin = 0;
const YDim_t yWorldMin = 0;
const XDim_t xWorldMax = 0x80000000;
const YDim_t yWorldMax = 0x80000000;

struct Point_t
{
  XDim_t x;
  YDim_t y;
  Point_t(XDim_t xx=0, YDim_t yy=0) :x(xx), y(yy) {}
};

struct Rect_t
{
  Point_t p1, p2;                       // p2 is not included in rectangle
  Rect_t(XDim_t x1=0, YDim_t y1=0, XDim_t x2=0, YDim_t y2=0)
    :p1(x1,y1), p2(x2,y2) {}
  Rect_t(Point_t pp1, Point_t pp2) :p1(pp1), p2(pp2) {}
  
  // returns true if rect contains point
  Boolean Contains(const Point_t &p)
  {
    return ((p.x >= p1.x) && (p.x < p2.x) &&
            (p.y >= p1.y) && (p.y < p2.y));
  }

  // returns the quadrant of rect in which a given point lies
  int QuadrantOf(const Point_t &p)
  {
    return ((p.y >= (p1.y+p2.y)/2)*2) + (p.x >= (p1.x+p2.x)/2);
  }
            
  //-----------------------------------------------------------
  //function to shrink the rectangle to the quadrant in which p
  //lies. Also returns the index of the quadrant. If p does not
  //lie in the rectangle, results are undefined.
  //-----------------------------------------------------------
  int Shrink(const Point_t &p)
  {
    int q = 0;
    Point_t mid((p1.x+p2.x)/2, (p1.y+p2.y)/2);
    *((p.y < mid.y)? &p2.y:(q+=2, &p1.y)) = mid.y;
    *((p.x < mid.x)? &p2.x:(q+=1, &p1.x)) = mid.x;
    return q;
  }
  // shrinks the rectangle to the given quadrant
  void Shrink(int q)
  {
    *((q & 2)? &p1.y: &p2.y) = (p1.y+p2.y)/2;
    *((q & 1)? &p1.x: &p2.x) = (p1.x+p2.x)/2;
  }
};

struct Data_t
{
  // Data goes in here..
};
struct Leaf_t
{
  XDim_t x;                             // cannot use Point_t since this
  YDim_t y;                             // struct is to be used in a union
  Data_t *data;
  unsigned _pad;                        // make it 16 bytes
};

enum PtrType_t {PTInternal, PTExternal};
const int PINil = -1;                   // a null page index (pointer)
const int SINil = -1;                   // a null slot index
                                        // these two MUST be equal to -1

struct Slot_t
{
  unsigned inuse:1;                     // is this slot in use
  unsigned isleaf:1;                    // is this a leaf node
  unsigned e0:1;                        // for each child, whether the 
  unsigned e1:1;                        // child points to an internal
  unsigned e2:1;                        // or an external node (if this
  unsigned e3:1;                        // is not a leaf node
  unsigned _filler:26;                  // padding
  union
  {
    int child[4];                       // four childern if not a leaf
    Leaf_t l;                           // data info, if leaf
  };
  Slot_t() :inuse(False) {}
  SetChildType(int child, PtrType_t type)
  {
    switch(child)
    {
    case 0: e0 = type; break;
    case 1: e1 = type; break;
    case 2: e2 = type; break;
    case 3: e3 = type; break;
    default: Assert(False);
    }
  }
  PtrType_t ChildType(int child)
  {
    switch(child)
    {
    case 0: return PtrType_t(e0);
    case 1: return PtrType_t(e1);
    case 2: return PtrType_t(e2);
    case 3: return PtrType_t(e3);
    default: Assert(False);
    }
  }
};
  
struct QPage_t
{
  short searchStart;                    // used for alloc
  short usedCount;                      // number of filled slots
  short root[4];                        // root locations for trees

  static const HdrSize = 12;            // CHANGE IF ADD MORE FIELDS ABOVE
  static const PageSize = (1<<12);      // 4K page
  static const SlotsPerPage = (PageSize-HdrSize)/sizeof(Slot_t);
  Slot_t slot[SlotsPerPage];            // the slots
  char _filler[PageSize-HdrSize-sizeof(Slot_t)*SlotsPerPage];
  
  QPage_t() 
    :searchStart(0), usedCount(0)
  {
    for (int i = 0; i < 4; root[i++] = SINil);
  }
  int TreeCount()
  {
    int count = 0;
    for (int i = 0; i < 4; count += (root[i++] != SINil)?1:0);
    return count;
  }
  
  int NodeCount(int rootSlot);
  int Traverse(int rootNum, Rect_t *pExtent, const Point_t &pt);
  int AllocSlot()
  {
    if (usedCount == SlotsPerPage)
      return SINil;
    int i = searchStart;
    for (int j = 0; (j < SlotsPerPage) && slot[i].inuse;
                    j++, i = (i+1) % SlotsPerPage);
    Assert(j != SlotsPerPage);          // there has to be a free slot
    usedCount += 1;
    slot[i].inuse = True;
    searchStart = (i+1) % SlotsPerPage;
    return i;
  }
  void FreeSlot(int i)
  {
    Assert((i>=0) && (i<SlotsPerPage));
    Assert(slot[i].inuse);
    slot[i].inuse = 0;
    usedCount -= 1;
  }
};
  

class Disk_t
{
  static const int MaxNoOfPages = (1<<18);
  QPage_t *pageArray[MaxNoOfPages];
  int searchStart;
public:
  Disk_t()
    :searchStart(0)
  {
    for (int i = 0; i < MaxNoOfPages; pageArray[i++] = NULL);
  }
  ~Disk_t()
  {
    for (int i = 0; i < MaxNoOfPages; i++)
      if (pageArray[i])
        delete pageArray[i];
  }
  int AllocPage()
  {
    int i = searchStart;
    for (int j = 0; (j < MaxNoOfPages) && (pageArray[i] != NULL); 
                    j++, i = (i+1) % MaxNoOfPages);
    if (j == MaxNoOfPages)
      return PINil;
    if ((pageArray[i] = new QPage_t) == NULL)
      return PINil;
    searchStart = (i+1) % MaxNoOfPages;
    return i;
  }
  QPage_t *FetchPage(int i)
  {
    Assert((0<=i) && (i < MaxNoOfPages));
    Assert(pageArray[i]);
    return pageArray[i];
  }
  int FreePage(int i)
  {
    Assert((0<=i) && (i < MaxNoOfPages));
    Assert(pageArray[i]);
    delete pageArray[i];
    pageArray[i] = NULL;
  }
  void PrintStats();
};

//----------------------------------------------------------------------
struct PathEntry_t
{
  int pageIndex;                        // index of this page
  QPage_t *page;                        // the page
  Rect_t rootExtent;                    // the rectangle bounding the root
  int qroot;                            // which root of the page
  Rect_t leafExtent;                    // the rectangle bounding the leaf
  int leafIndex;                        // the slot which cause exit from
                                        // page
};

//----------------------------------------------------------------------
typedef void (*NodeOp_t)(Slot_t &s, int nodeLevel, int pageLevel);
typedef void (*LeafOp_t)(Leaf_t &l, int nodeLevel, int pageLevel);
struct QTree_t
{
  int rootPage;
  Rect_t extent;
  
  QTree_t() :rootPage(PINil), extent(xWorldMin, yWorldMin, xWorldMax,
                                  yWorldMax) {}
  Boolean Insert(const Point_t &pt, Data_t *data);
  Data_t *Remove(const Point_t &pt);
  Data_t *Search(const Point_t &pt);
  void Compact(int M);
  void Print();
  void Traverse(LeafOp_t fn); 
  void PrintStats();
private:
  int BuildPath(PathEntry_t path[], const Point_t &pt);
  void RecTraverse(QPage_t *page, int si, LeafOp_t fn, 
                   int level, int pageLevel);
};

// ----------------------------------------------------------------------
// COMPACTION Related in memory representation of a qnode structure
// ----------------------------------------------------------------------

struct QTNode_t
{
  unsigned isleaf:1;                    // is this a leaf node
  unsigned e0:1;                        
  unsigned e1:1 ;
  unsigned e2:1 ;
  unsigned e3:1 ;

  QTNode_t* child[4];                   // four in memory children if not a leaf
  int page[4] ;			// four child in other page 
  Leaf_t l;                           // data info, if leaf
  

  QTNode_t()  {
  
  	for ( int q = 0 ; q < 4 ; child[q] = NULL,page[q++] = -1 ) ;

  }		

  // PtrType_t PTRInternal means that the node is in memory 	
  SetChildType(int child, PtrType_t type)
  {
    switch(child)
    {
    case 0: e0 = type; break;
    case 1: e1 = type; break;
    case 2: e2 = type; break;
    case 3: e3 = type; break;
    default: Assert(False);
    }
  }

  PtrType_t ChildType(int child)
  {
    switch(child)
    {
    case 0: return PtrType_t(e0);
    case 1: return PtrType_t(e1);
    case 2: return PtrType_t(e2);
    case 3: return PtrType_t(e3);
    default: Assert(False);
    }
  }
};

extern Disk_t disk;
#endif /* __QPAGE_H */
